/*
Sonar configuration
*/
int sonarTriggerPin         = D1;
int sonarEchoPin            = D2;
long sonarPingInterval      = 200;  // ms
long sonarHoverDelay        = 500;  // ms
long sonarDistanceThreshold = 500;  // mm

#define EMPTY 1
#define ENTERING 2
#define PRESENT 3
#define LEAVING 4

/*
Servo configuration
*/
int servoDirectionPin       = D0;
long servoStepDuration      = 100;  // ms
long servoWindingDuration   = 4000; // ms

#define WINDING_UP 180
#define WINDING_DOWN 0
#define IDLE 90

    
class Sonar {
    int sonarState;
    unsigned long previousHoverMillis;
    unsigned long previousPingMillis;

    private:
        long getDistance_mm() {
            digitalWrite(sonarTriggerPin, HIGH);
            delayMicroseconds(50);
            digitalWrite(sonarTriggerPin, LOW);
            long Time_Echo_us = pulseIn(sonarEchoPin, HIGH);
            if((Time_Echo_us < 60000) && (Time_Echo_us > 10)) {
              long distance_mm = (Time_Echo_us*34/100)/2;
              return distance_mm;
            }
            return 1000;
        }
    public:
        Sonar(int a) {
            sonarState = EMPTY; 
            previousHoverMillis = 0;
            previousPingMillis = 0;
            pinMode(sonarTriggerPin, OUTPUT); 
            pinMode(sonarEchoPin, INPUT); 
        }
        bool Update(bool& state) {
            unsigned long currentMillis = millis();
            bool isDetected;
            if (currentMillis - previousPingMillis > sonarPingInterval) {
                isDetected = getDistance_mm() < sonarDistanceThreshold;
                
                if (isDetected) {
                    if (sonarState == EMPTY) sonarState = ENTERING;
                    if (sonarState == LEAVING) sonarState = PRESENT;
                } else {
                    if (sonarState == PRESENT) sonarState = LEAVING;
                    if (sonarState == ENTERING) sonarState = EMPTY;
                }
                previousPingMillis = currentMillis;
            }
            if (currentMillis - previousHoverMillis > sonarHoverDelay) {
                if (sonarState == ENTERING && isDetected) sonarState = PRESENT;
                if (sonarState == LEAVING && !isDetected) sonarState = EMPTY;
                previousHoverMillis = currentMillis;
            }
            return (sonarState == PRESENT || sonarState == LEAVING);
         }
};



class ServoWinder {
    int servoStatus;
    unsigned long stepsThresholdUp;
    unsigned long stepsThresholdDown;
    unsigned long previousServoMillis;
    
    unsigned long currentSteps;
    Servo microServo; 
    
    private:
        bool isUpReached() {
            return currentSteps >= stepsThresholdUp;
        }
        bool isDownReached() {
            return currentSteps <= stepsThresholdDown;
        }
        int getDirection(bool isPresent) {
            if((isPresent || servoStatus == WINDING_UP) && !isUpReached()) {
               return WINDING_UP;
            } else if((!isPresent || servoStatus == WINDING_DOWN)  && !isDownReached()) {
                return WINDING_DOWN;
            } else {
                return IDLE;
            }
        }
        void setServoDirection(int direction) {
            microServo.write(direction);
            if(direction == WINDING_UP) currentSteps += servoStepDuration;
            if(direction == WINDING_DOWN) currentSteps -= servoStepDuration;
        }
    public:
        ServoWinder(int seed) {
            stepsThresholdUp = servoWindingDuration;
            stepsThresholdDown = 0;
            currentSteps = 0;
            previousServoMillis = 0;
            servoStatus = IDLE;

            microServo.attach(servoDirectionPin);
        }
        void Update(bool isPresent) {
            unsigned long currentMillis = millis();
            if (currentMillis - previousServoMillis > servoStepDuration) {
                servoStatus = getDirection(isPresent);
                setServoDirection(servoStatus);
                previousServoMillis = currentMillis;
            }
        }
};


/*
Init & main loop
*/

Sonar sonar();
ServoWinder servo();

void setup() {
}
void loop() {
    bool sonarState = sonar.Update(sonarState);
    servo.Update(sonarState);
}